from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("A")
class Electric_Current(Unit): pass

@Unitor("A",1)
class A(Electric_Current): pass
@Unitor("mA",10 ** -3)
class MA(Electric_Current): pass


from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("v")
class Speed(Unit): pass

# METRIC
@Unitor("km/h",1/3.6)
class KMH(Speed): pass
@Unitor("m/s",1)
class MS(Speed): pass

# IMPERIAl
@Unitor("mph",1/3.6 * 1.609344)
class MPH(Speed): pass
@Unitor("fps",1/3.6 * 1.09728)
class FPS(Speed): pass

# SPACE
@Unitor("c",1/3.6 * 1.07925285 * (10 ** 9))
class LightSpeed(Speed): pass



from typing import overload
from uniter.Uniter import Unit
from uniter.units import *

@overload
def average_speed(distance: None = ..., speed: Speed = ..., time: Time = ...) -> Length: ...

@overload
def average_speed(distance: Length = ..., speed: None = ..., time: Time = ...) -> Speed: ...

@overload
def average_speed(distance: Length = ..., speed: Speed = ..., time: None = ...) -> Time: ...

def average_speed(distance: Length | None = ..., speed: Speed | None = ..., time: Time | None = ...) -> Unit: ...

@overload
def cylinder_surface(radius: None = ..., height: Length = ..., surface: Area = ...) -> Length: ...

@overload
def cylinder_surface(radius: Length = ..., height: None = ..., surface: Area = ...) -> Length: ...

@overload
def cylinder_surface(radius: Length = ..., height: Length = ..., surface: None = ...) -> Area: ...

def cylinder_surface(radius: Length | None = ..., height: Length | None = ..., surface: Area | None = ...) -> Unit: ...

@overload
def cylinder_volume(radius: None = ..., height: Length = ..., volume: Volume = ...) -> Length: ...

@overload
def cylinder_volume(radius: Length = ..., height: None = ..., volume: Volume = ...) -> Length: ...

@overload
def cylinder_volume(radius: Length = ..., height: Length = ..., volume: None = ...) -> Volume: ...

def cylinder_volume(radius: Length | None = ..., height: Length | None = ..., volume: Volume | None = ...) -> Unit: ...

@overload
def density_formula(mass: None = ..., volume: Volume = ..., density: Density = ...) -> Mass: ...

@overload
def density_formula(mass: Mass = ..., volume: None = ..., density: Density = ...) -> Volume: ...

@overload
def density_formula(mass: Mass = ..., volume: Volume = ..., density: None = ...) -> Density: ...

def density_formula(mass: Mass | None = ..., volume: Volume | None = ..., density: Density | None = ...) -> Unit: ...

@overload
def ohms_law(electric_current: None = ..., voltage: Voltage = ..., resistance: Resistance = ...) -> Electric_Current: ...

@overload
def ohms_law(electric_current: Electric_Current = ..., voltage: None = ..., resistance: Resistance = ...) -> Voltage: ...

@overload
def ohms_law(electric_current: Electric_Current = ..., voltage: Voltage = ..., resistance: None = ...) -> Resistance: ...

def ohms_law(electric_current: Electric_Current | None = ..., voltage: Voltage | None = ..., resistance: Resistance | None = ...) -> Unit: ...

@overload
def sphere_surface(radius: None = ..., surface: Area = ...) -> Length: ...

@overload
def sphere_surface(radius: Length = ..., surface: None = ...) -> Area: ...

def sphere_surface(radius: Length | None = ..., surface: Area | None = ...) -> Unit: ...

@overload
def sphere_volume(radius: None = ..., volume: Volume = ...) -> Length: ...

@overload
def sphere_volume(radius: Length = ..., volume: None = ...) -> Volume: ...

def sphere_volume(radius: Length | None = ..., volume: Volume | None = ...) -> Unit: ...


from uniter.Uniter import Unit, __Q

class Speed(Unit[__Q]): pass

class FPS(Speed[Speed]): pass

class KMH(Speed[Speed]): pass

class LightSpeed(Speed[Speed]): pass

class MPH(Speed[Speed]): pass

class MS(Speed[Speed]): pass




from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("°")
class Angle(Unit):
    def __conv__(self, unit):
        from math import degrees, radians
        DEGS = [DEG, MOA, SOA]
        if self.__class__ in DEGS and unit in DEGS:
            return super().__conv__(unit) # type: ignore
        elif self.__class__ in DEGS and unit is RAD:
            return radians(float(self[DEG]))
        elif self.__class__ is RAD and unit in DEGS:
            return float(DEG(degrees(float(self)))[unit])


@Unitor("°", 1)
class DEG(Angle): pass

@Unitor("′", 1/60)
class MOA(Angle): pass

@Unitor("″", 1/3600)
class SOA(Angle): pass

@Unitor("rad")
class RAD(Angle): pass



from uniter.Uniter import Unit, __Q

class Computer_Memory(Unit[__Q]): pass

class Bit(Computer_Memory[Computer_Memory]): pass

class Byte(Computer_Memory[Computer_Memory]): pass

class EB(Computer_Memory[Computer_Memory]): pass

class Eb(Computer_Memory[Computer_Memory]): pass

class GB(Computer_Memory[Computer_Memory]): pass

class Gb(Computer_Memory[Computer_Memory]): pass

class KB(Computer_Memory[Computer_Memory]): pass

class Kb(Computer_Memory[Computer_Memory]): pass

class MB(Computer_Memory[Computer_Memory]): pass

class Mb(Computer_Memory[Computer_Memory]): pass

class PB(Computer_Memory[Computer_Memory]): pass

class Pb(Computer_Memory[Computer_Memory]): pass

class TB(Computer_Memory[Computer_Memory]): pass

class Tb(Computer_Memory[Computer_Memory]): pass

class YB(Computer_Memory[Computer_Memory]): pass

class Yb(Computer_Memory[Computer_Memory]): pass

class ZB(Computer_Memory[Computer_Memory]): pass

class Zb(Computer_Memory[Computer_Memory]): pass




from uniter.Uniter import Unit
from uniter.units import *
from uniter.Exceptions import UnitFormulaParameterError
from sympy import Eq, solve
from sympy.core.symbol import Symbol


def __form__(eq, unit_map, **values):
    decl_Size = {k: v for k, v in values.items() if v is not None}.__len__()
    if decl_Size < values.__len__() - 1:
        raise UnitFormulaParameterError("You did not provide enough parameters to calculation!")
    if decl_Size == values.__len__():
        raise UnitFormulaParameterError("You provide all arguments, and there is nothing to calculate!")
    for s in eq.free_symbols:
        val, un = values[str(s)], unit_map[str(s)]
        if val is not None:
            if not isinstance(val, Unit):
                raise UnitFormulaParameterError(
                    f"Expected unit of {un.__base__.__name__}, got {type(val).__name__} instead")
            if not val.is_convertable_to(un):
                raise UnitFormulaParameterError(
                    f"Expected unit of {un.__base__.__name__}, got unit of {type(val).__base__.__name__} instead")
            eq = eq.subs(s, float(val[un]))
    runit, res = list(solve(eq, dict=True)[0].items())[0]
    return unit_map[str(runit)](res)


def ohms_law(electric_current: Electric_Current = None, voltage: Voltage = None, resistance: Resistance = None):
    return __form__(Eq(Symbol("i"), Symbol("u") / Symbol("r")), {"r": OHM, "u": Volt, "i": A}, i=electric_current,
                    u=voltage, r=resistance)


def average_speed(distance: Length = None, speed: Speed = None, time: Time = None):
    return __form__(Eq(Symbol("s"), Symbol("v") * Symbol("t")), {"s": M, "v": MS, "t": SEC}, s=distance, v=speed,
                    t=time)


def cylinder_volume(radius: Length = None, height: Length = None, volume: Volume = None):
    from math import pi
    return __form__(Eq(pi * Symbol("r") ** 2 * Symbol("v"), Symbol("V")), {"r": M, "v": M, "V": M_3}, r=radius,
                    v=height, V=volume)


def cylinder_surface(radius: Length = None, height: Length = None, surface: Area = None):
    from math import pi
    return __form__(Eq(2 * pi * Symbol("r") * (Symbol("r") + Symbol("v")), Symbol("S")), {"r": M, "v": M, "S": M_2},
                    r=radius, v=height, S=surface)


def sphere_volume(radius: Length = None, volume: Volume = None):
    from math import pi
    return __form__(Eq(4 / 3 * pi * Symbol("r") ** 3, Symbol("V")), {"r": M, "V": M_3}, r=radius, V=volume)


def sphere_surface(radius: Length = None, surface: Area = None):
    from math import pi
    return __form__(Eq(4 * pi * Symbol("r") ** 2), {"r": M, "S": M_2}, r=radius, S=surface)


def density_formula(mass: Mass = None, volume: Volume = None, density: Density = None):
    return __form__(Eq(Symbol("R"), Symbol("m") / Symbol("V")), {"R": KGM_3, "m": KG, "V": M_3}, m=mass, V=volume,
                    R=density)



from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("V")
class Volume(Unit): pass

@Unitor("m^3", 10**3)
class M_3(Volume): pass

@Unitor("hL", 10**2)
class HL(Volume): pass

@Unitor("L", 1)
class L(Volume): pass

@Unitor("dm^3", 1)
class DM_3(Volume): pass

@Unitor("dL", 10**-1)
class DL(Volume): pass

@Unitor("cL", 10**-2)
class CL(Volume): pass

@Unitor("mL", 10**-3)
class ML(Volume): pass

@Unitor("cm^3", 10**-3)
class CM_3(Volume): pass


"""
Basic package to handle unit conversion. It includes all units, even those which are not currently using.
"""

from .Uniter import UnitType, get_all_quantities, get_all_units
from .Parser import parse

from .units import *


from uniter.Uniter import Unit, __Q

class Time(Unit[__Q]): pass

class DAY(Time[Time]): pass

class HR(Time[Time]): pass

class MIN(Time[Time]): pass

class SEC(Time[Time]): pass

class YR(Time[Time]): pass




from uniter.Uniter import Unit,Unitor,Quantitor

@Quantitor()
class Computer_Memory(Unit): pass

@Unitor("b",1)
class Bit(Computer_Memory): pass

@Unitor("B",8)
class Byte(Computer_Memory): pass

@Unitor("Kb",2 ** 10)
class Kb(Computer_Memory): pass
@Unitor("KB",8 * 2 ** 10)
class KB(Computer_Memory): pass

@Unitor("Mb",2 ** 20)
class Mb(Computer_Memory): pass
@Unitor("MB",8 * 2 ** 20)
class MB(Computer_Memory): pass

@Unitor("Gb",2 ** 30)
class Gb(Computer_Memory): pass
@Unitor("GB",8 * 2 ** 30)
class GB(Computer_Memory): pass

@Unitor("Tb",2 ** 40)
class Tb(Computer_Memory): pass
@Unitor("TB",8 * 2 ** 40)
class TB(Computer_Memory): pass

@Unitor("Pb",2 ** 50)
class Pb(Computer_Memory): pass
@Unitor("PB",8 * 2 ** 50)
class PB(Computer_Memory): pass

@Unitor("Eb",2 ** 60)
class Eb(Computer_Memory): pass
@Unitor("EB",8 * 2 ** 60)
class EB(Computer_Memory): pass

@Unitor("Zb",2 ** 70)
class Zb(Computer_Memory): pass
@Unitor("ZB",8 * 2 ** 70)
class ZB(Computer_Memory): pass

@Unitor("Yb",2 ** 80)
class Yb(Computer_Memory): pass
@Unitor("YB",8 * 2 ** 80)
class YB(Computer_Memory): pass


from uniter.Uniter import Unit, Unitor, Quantitor


@Quantitor("R")
class Resistance(Unit): pass


@Unitor("Ω", 1)
class OHM(Resistance): pass


@Unitor("kΩ", 10 ** 3)
class KOHM(Resistance): pass


@Unitor("MΩ", 10 ** 6)
class MOHM(Resistance): pass



from uniter.Uniter import Unit, __Q

class Volume(Unit[__Q]): pass

class CL(Volume[Volume]): pass

class CM_3(Volume[Volume]): pass

class DL(Volume[Volume]): pass

class DM_3(Volume[Volume]): pass

class HL(Volume[Volume]): pass

class L(Volume[Volume]): pass

class ML(Volume[Volume]): pass

class M_3(Volume[Volume]): pass




from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("S")
class Area(Unit): pass

@Unitor("km^2", 10**6)
class KM_2(Area): pass

@Unitor("ha", 10**4)
class HA(Area): pass

@Unitor("are", 10**2)
class ARE(Area): pass

@Unitor("m^2", 1)
class M_2(Area): pass

@Unitor("m^2", 10**-2)
class DM_2(Area): pass

@Unitor("m^2", 10**-4)
class CM_2(Area): pass

@Unitor("m^2", 10**-6)
class MM_2(Area): pass


from uniter.Uniter import Unit, Unitor, Quantitor

@Quantitor("t")
class Time(Unit): pass

@Unitor("sec",1)
class SEC(Time): pass
@Unitor("min",60)
class MIN(Time): pass
@Unitor("h",60 ** 2)
class HR(Time): pass
@Unitor("d",60 ** 2 * 24)
class DAY(Time): pass
@Unitor("yr",60 ** 2 * 24 * 365.242199)
class YR(Time): pass


class default_exception_module:
    def __call__(self, cls):
        cls.__module__ = Exception.__module__
        return cls


@default_exception_module()
class UnitInstatiateError(Exception): pass

@default_exception_module()
class UnitConversionError(Exception): pass

@default_exception_module()
class UnitArithmeticError(Exception): pass

@default_exception_module()
class UnitInheritanceError(Exception): pass

@default_exception_module()
class UnitBaseClassError(Exception): pass

@default_exception_module()
class UnitQuantitorError(Exception): pass

@default_exception_module()
class UnitUnitorError(Exception): pass

@default_exception_module()
class UnknownUnitError(Exception): pass

@default_exception_module()
class UnitFormulaParameterError(Exception): pass





from uniter.Uniter import Unit, __Q

class Mass(Unit[__Q]): pass

class G(Mass[Mass]): pass

class KG(Mass[Mass]): pass

class MG(Mass[Mass]): pass

class Ounce(Mass[Mass]): pass

class Pound(Mass[Mass]): pass

class Stone(Mass[Mass]): pass

class T(Mass[Mass]): pass




from uniter.Uniter import Unit, __Q

class Area(Unit[__Q]): pass

class ARE(Area[Area]): pass

class CM_2(Area[Area]): pass

class DM_2(Area[Area]): pass

class HA(Area[Area]): pass

class KM_2(Area[Area]): pass

class MM_2(Area[Area]): pass

class M_2(Area[Area]): pass




from uniter.Uniter import Unit, Unitor, UnitType, Quantitor

@Quantitor("m")
class Mass(Unit): pass

# METRIC
@Unitor("t",10 ** 6, UnitType.METRIC)
class T(Mass): pass
@Unitor("kg",10 ** 3, UnitType.METRIC)
class KG(Mass): pass
@Unitor("g",1, UnitType.METRIC)
class G(Mass): pass
@Unitor("mg",10 ** -3, UnitType.METRIC)
class MG(Mass): pass

# IMPERIAL
@Unitor("lb", 10 ** 3 * 0.45359237, UnitType.IMPERIAL)
class Pound(Mass): pass
@Unitor("lb", 28.349523125, UnitType.IMPERIAL)
class Ounce(Mass): pass
@Unitor("st", 10 ** 3 * 6.35029318, UnitType.IMPERIAL)
class Stone(Mass): pass


from uniter.Uniter import Unit, __Q

class Temperature(Unit[__Q]): pass

class DEG_C(Temperature[Temperature]): pass

class DEG_F(Temperature[Temperature]): pass

class DEG_K(Temperature[Temperature]): pass




from uniter.Uniter import Unit, Unitor, Quantitor


@Quantitor("T")
class Temperature(Unit):
    def __conv__(self, unit):
        from sympy import symbols, Eq, solve
        C, K, F = symbols("C K F")
        SYM = {DEG_C: C, DEG_K: K, DEG_F: F}
        EQ = {(DEG_C, DEG_K): Eq(C + 273.15, K), (DEG_C, DEG_F): Eq(9 / 5 * C + 32, F),
              (DEG_K, DEG_F): Eq(9 / 5 * K - 459.67, F)}

        return solve(
            EQ[[k for k in EQ.keys() if set(k) == {self.__class__, unit}][0]].subs(SYM[self.__class__], float(self)) # type: ignore
        )[0]


@Unitor("°C", 1)
class DEG_C(Temperature): pass


@Unitor("°F")
class DEG_F(Temperature): pass


@Unitor("°K")
class DEG_K(Temperature): pass



from .Exceptions import UnknownUnitError
from .units import *

def parse(expression, unit_mapping = None):
    import re
    from .Uniter import Unit
    MAP = {} if unit_mapping is None else {k:v.__name__ for k, v in unit_mapping.items()}
    for q in Unit.__subclasses__():
        for u in q.__subclasses__():
            MAP[u.symbol] = u.__name__
    for match in re.finditer(r"(\d+)(?: |)+(\w+)",expression):
        if match.group(2) not in MAP:
            raise UnknownUnitError(f"Unknown unit \"{match.group(2)}\"")
        expression = expression.replace(match.group(0),f"{MAP[match.group(2)]}({match.group(1)})")
    return eval(expression)


from typing import TypeVar, Generic, Type

__Q = TypeVar("__Q", bound=Unit)


class Unit(Generic[__Q]):
    __value: float
    __multiplier: float
    __symbol: str
    __unit_type: UnitType
    __sign: str

    @property
    def multiplier(self) -> float: ...

    @property
    def symbol(self) -> str: ...

    @property
    def unit_type(self) -> UnitType: ...

    @property
    def sign(self) -> UnitType: ...

    def convert_to(self, unit: type[__Q]) -> __Q: ...

    def in_bigger_unit(self, keep_unit_type: bool = ...) -> __Q: ...

    def in_smaller_unit(self, keep_unit_type: bool = ...) -> __Q: ...

    def in_default_unit(self) -> __Q: ...

    def is_convertable_to(self, unit: type[__Q]) -> bool: ...

    @classmethod
    def default_unit(cls) -> type[__Q]: ...

    @classmethod
    def units_by_category(cls, unit_type: UnitType) -> list[__Q]: ...

    @classmethod
    def bigger_unit(cls, keep_unit_type: bool = ...) -> Type[__Q]: ...

    @classmethod
    def smaller_unit(cls, keep_unit_type: bool = ...) -> Type[__Q]: ...

    @classmethod
    def is_biggest(cls, in_unit_type: bool = ...) -> bool: ...

    @classmethod
    def is_smallest(cls, in_unit_type: bool = ...) -> bool: ...

    def __init__(self, value: int | float) -> __Q: ...

    def __add__(self, other: __Q) -> __Q: ...

    def __sub__(self, other: __Q) -> __Q: ...

    def __getitem__(self, item: type[__Q]) -> __Q: ...

    def __mul__(self, other: int | float) -> __Q: ...

    def __rmul__(self, other: int | float) -> __Q: ...

    def __pow__(self, power: int | float, modulo=...) -> __Q: ...

    def __truediv__(self, other: int | float) -> __Q: ...

    def __floordiv__(self, other: int | float) -> __Q: ...

    def __str__(self) -> str: ...

    def __repr__(self) -> str: ...

    def __int__(self) -> int: ...

    def __float__(self) -> float: ...

    def __eq__(self, other: __Q) -> bool: ...

    def __ne__(self, other: __Q) -> bool: ...

    def __lt__(self, other: __Q) -> bool: ...

    def __le__(self, other: __Q) -> bool: ...

    def __gt__(self, other: __Q) -> bool: ...

    def __ge__(self, other: __Q) -> bool: ...


class Unitor:
    __mp: float
    __sy: str
    __ut: UnitType

    def __init__(self, symbol: str = ..., mp: int | float = ..., unit_type: UnitType = ...): ...

    def __call__(self): ...


class Quantitor:
    __s: str

    def __init__(self, sign: str | None = ...): ...

    def __call__(self): ...


class UnitType:
    METRIC = ...
    IMPERIAL = ...
    ASTRONOMICAL = ...

def get_all_quantities() -> list[type[Unit]]: ...

def get_all_units() -> list[type[Unit]]: ...


from uniter.Uniter import Unit, Quantitor, Unitor

@Quantitor("W")
class Energy(Unit): pass

@Unitor("J", 1)
class J(Energy): pass

@Unitor("kJ", 10 ** 3)
class KJ(Energy): pass

@Unitor("MJ", 10 ** 6)
class MJ(Energy): pass

@Unitor("GJ", 10 ** 9)
class GJ(Energy): pass


from uniter.Uniter import Unit, __Q

class Length(Unit[__Q]): pass

class AstronomicalUnit(Length[Length]): pass

class CM(Length[Length]): pass

class DM(Length[Length]): pass

class Foot(Length[Length]): pass

class Inch(Length[Length]): pass

class KM(Length[Length]): pass

class LightYear(Length[Length]): pass

class M(Length[Length]): pass

class MM(Length[Length]): pass

class McM(Length[Length]): pass

class Mile(Length[Length]): pass

class NM(Length[Length]): pass

class PM(Length[Length]): pass

class Yard(Length[Length]): pass




from enum import Enum

from .Exceptions import UnitInstatiateError, UnitConversionError, UnitInheritanceError, UnitQuantitorError, \
    UnitUnitorError, UnitArithmeticError, UnitBaseClassError
from .Util import classproperty


class Unit:

    def __init__(self, value):
        if type(self) is Unit:
            raise UnitInstatiateError("Cannot instantiate Unit base class")
        if type(self).__base__ is Unit:
            raise UnitInstatiateError(f"Cannot instatiate {type(self).__name__} quantity class")
        if type(self).__base__.__base__ is not Unit:
            raise UnitInstatiateError(
                f"Cannot instatiate {type(self).__name__} class that is not child of quantity class")
        if self.__class__.__dict__.__len__() == 2:
            self.__class__.__multiplier = 0
            self.__class__.__symbol = ""
            self.__class__.__unit_type = None
            self.__class__.__sign = ""
        self.__value = value

    @classproperty
    def multiplier(self) -> float:
        return self.__multiplier

    @classproperty
    def symbol(self):
        return self.__symbol

    @classproperty
    def unit_type(self):
        return self.__unit_type

    @classproperty
    def sign(self):
        return self.__sign

    def convert_to(self, unit):
        if type(self) is unit:
            return self
        if type(unit) is not type or not isinstance(self, unit.__base__):
            sbs = self.__class__.__base__.__subclasses__()
            cvt = unit.__base__.__name__ if type(unit) is type else type(unit).__name__
            raise UnitConversionError(
                f"Illegal conversion from {self.__class__.__base__.__name__} to {cvt} available units to convert this object to: {', '.join([c.__name__ for c in sbs])}")

        return unit(self.__conv__(unit))  # type: ignore

    def in_bigger_unit(self, keep_unit_type=True):
        return self[self.__class__.bigger_unit(keep_unit_type)]

    def in_smaller_unit(self, keep_unit_type=True):
        return self[self.__class__.smaller_unit(keep_unit_type)]

    def in_default_unit(self):
        return self[self.default_unit()]

    def is_convertable_to(self, unit):
        try:
            self[unit]
        except UnitConversionError:
            return False
        return True

    @classmethod
    def bigger_unit(cls, keep_unit_type=True):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get bigger unit of Unit base class!")
        if cls.multiplier == 0: return cls
        units, index = cls.__unit_size_map__(keep_unit_type, "bigger unit")  # type: ignore
        return units[min(index + 1, units.__len__() - 1)]

    @classmethod
    def smaller_unit(cls, keep_unit_type=True):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get the smaller of Unit base class!")
        if cls.multiplier == 0: return cls
        units, index = cls.__unit_size_map__(keep_unit_type, "smaller unit")  # type: ignore
        return units[max(index - 1, 0)]

    @classmethod
    def is_biggest(cls, in_unit_type=True):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get the biggest unit of Unit base class!")
        if cls.multiplier == 0: return False
        units, index = cls.__unit_size_map__(in_unit_type, "if unit is the biggest")  # type: ignore
        return index == units.__len__() - 1

    @classmethod
    def is_smallest(cls, in_unit_type=True):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get the smallest unit of Unit base class!")
        if cls.multiplier == 0: return False
        units, index = cls.__unit_size_map__(in_unit_type, "if unit is the smallest")  # type: ignore
        return index == 0

    @classmethod
    def default_unit(cls):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get default method of Unit base class!")
        clss = (cls if cls.__base__ is Unit else cls.__base__).__subclasses__()
        return [c for c in clss if c.multiplier == 1][0]  # type: ignore

    @classmethod
    def units_by_category(cls, unit):
        if cls is Unit:
            raise UnitBaseClassError("Cannot get units from Unit base class!")
        clss = (cls if cls.__base__ is Unit else cls.__base__).__subclasses__()
        return [c for c in clss if c.unit_type is unit]

    @classmethod
    def __unit_size_map__(cls, keep_unit_type, operation_name):
        if cls is Unit or cls.__base__ is Unit:
            raise UnitBaseClassError(
                f"Cannot get {operation_name} from {'Unit base class' if cls is Unit else f'{cls.__name__} quantity class'}")
        if cls.__base__.__base__ is not Unit:
            raise UnitInheritanceError(
                f"Cannot get {operation_name} from {cls.__name__} class because it is not in correct inheritance order (Unit -> QuantityName -> {cls.__name__})")
        if keep_unit_type:
            uts = [c for c in cls.units_by_category(cls.unit_type) if c.multiplier != 0]  # type: ignore
        else:
            uts = [c for c in cls.__base__.__subclasses__() if c.multiplier != 0]  # type: ignore
        units = sorted(uts, key=lambda o: o.multiplier)
        index = units.index(cls)
        return units, index

    @staticmethod
    def __clr_pd__(num):
        from re import sub
        return float(sub(r"(\.\d+?)(0)(\2+)$", "\1", str(num)))

    def __conv__(self, unit):
        return self.__value * self.__multiplier / unit(0).multiplier

    def __calc__(self, other, oper_name, oper_symbol):
        op = {'+': lambda x, y: x + y, '-': lambda x, y: x - y}
        if self.__class__.__base__ is not other.__class__.__base__:
            raise UnitArithmeticError(
                f"{oper_name} of non-equal units ({self.__class__.__base__.__name__} {oper_symbol} {other.__class__.__base__.__name__})")
        return other.__class__(op[oper_symbol](self.__conv__(other.__class__), other.__value))

    def __add__(self, other):
        return self.__calc__(other, "Addition", "+")  # type: ignore

    def __sub__(self, other):
        return self.__calc__(other, "Subtraction", "-")  # type: ignore

    def __getitem__(self, unit):
        return self.convert_to(unit)

    def __mul__(self, other):
        if not isinstance(other, (int, float)):
            raise UnitArithmeticError(f"Multiplication of Unit with {other.__class__.__name__}, use int/float instead!")
        self.__value *= other
        return self

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, power, modulo=None):
        if not isinstance(power, (int, float)):
            raise UnitArithmeticError(f"Power of Unit with {power.__class__.__name__}, use int/float instead!")
        self.__value = pow(self.__value, power, modulo)
        return self

    def __truediv__(self, other):
        if not isinstance(other, (int, float)):
            raise UnitArithmeticError(f"Division of Unit with {other.__class__.__name__}, use int/float instead!")
        self.__value /= other
        return self

    def __floordiv__(self, other):
        if not isinstance(other, (int, float)):
            raise UnitArithmeticError(f"Division of Unit with {other.__class__.__name__}, use int/float instead!")
        self.__value //= other
        return self

    def __str__(self):
        return f"{float(self) if round(self.__value) != self.__value else int(self.__value)}{self.symbol}"

    def __repr__(self):
        return f"<{self.__class__.__base__.__name__}.{self.__class__.__name__} value={self.__value}>"

    def __int__(self):
        return round(self.__value)

    def __float__(self):
        return self.__clr_pd__(float(self.__value))  # type: ignore

    def __logic__(self, other, logiclambda):
        if not isinstance(other, self.__class__.__base__): return False
        return logiclambda(self[other.__class__].__value, other.__value)  # type: ignore

    def __eq__(self, other):
        return self.__logic__(other, lambda a, b: a == b)  # type: ignore

    def __ne__(self, other):
        return self.__logic__(other, lambda a, b: a != b)  # type: ignore

    def __lt__(self, other):
        return self.__logic__(other, lambda a, b: a < b)  # type: ignore

    def __le__(self, other):
        return self.__logic__(other, lambda a, b: a <= b)  # type: ignore

    def __gt__(self, other):
        return self.__logic__(other, lambda a, b: a > b)  # type: ignore

    def __ge__(self, other):
        return self.__logic__(other, lambda a, b: a >= b)  # type: ignore


class UnitType(Enum):
    METRIC = 0
    IMPERIAL = 1
    ASTRONOMICAL = 2


class Unitor:

    def __init__(self, symbol="", mp=0, unit_type=None) -> None:
        self.__mp = mp
        self.__sy = symbol
        self.__ut = unit_type

    def __call__(self, cls: type):
        if type(cls) is not type:
            raise UnitUnitorError("Unitor cannot be applied on function")
        if cls.__base__.__base__ is not Unit:
            raise UnitUnitorError("Unitor class is not extending Quantity class")
        cls._Unit__multiplier = self.__mp
        cls._Unit__symbol = self.__sy
        cls._Unit__unit_type = self.__ut
        return cls


class Quantitor:

    def __init__(self, sign=None):
        self.__s = sign

    def __call__(self, cls):
        if type(cls) is not type:
            raise UnitQuantitorError("Quantitior cannot be applied on function")
        if cls.__base__ is not Unit:
            raise UnitQuantitorError("Quantitor class is not extending Unit class")
        cls._Unit__sign = self.__s
        return cls


def get_all_quantities():
    return Unit.__subclasses__()


def get_all_units():
    r = []
    for q in get_all_quantities(): r.extend(q.__subclasses__())
    return r



from uniter.Uniter import Unit, Unitor, UnitType, Quantitor

@Quantitor("l")
class Length(Unit): pass

# METRIC
@Unitor("km",10 ** 3, UnitType.METRIC)
class KM(Length): pass
@Unitor("m",1, UnitType.METRIC)
class M(Length): pass
@Unitor("dm",10 ** -1, UnitType.METRIC)
class DM(Length): pass
@Unitor("cm",10 ** -2, UnitType.METRIC)
class CM(Length): pass
@Unitor("mm",10 ** -3, UnitType.METRIC)
class MM(Length): pass
@Unitor("µm",10 ** -6, UnitType.METRIC)
class McM(Length): pass
@Unitor("nm",10 ** -9, UnitType.METRIC)
class NM(Length): pass
@Unitor("pm",10 ** -12, UnitType.METRIC)
class PM(Length): pass

# IMPERIAL
@Unitor("mi",10 ** 3 * 1.609344, UnitType.IMPERIAL)
class Mile(Length): pass
@Unitor("yd",0.9144, UnitType.IMPERIAL)
class Yard(Length): pass
@Unitor("ft",10 ** -2 * 30.48, UnitType.IMPERIAL)
class Foot(Length): pass
@Unitor("in",10 ** -2 * 2.54, UnitType.IMPERIAL)
class Inch(Length): pass

# SPACE
@Unitor("au",10 ** 3 * 149597871, UnitType.ASTRONOMICAL)
class AstronomicalUnit(Length): pass
@Unitor("ly",10 ** 15 * 9.4605284, UnitType.ASTRONOMICAL)
class LightYear(Length): pass


from .Length import Length, KM, M, DM, CM, MM, McM, NM, PM, Mile, Yard, Foot, Inch, AstronomicalUnit, LightYear
from .Mass import Mass, T, KG, G, MG, Pound, Ounce, Stone
from .Time import Time, SEC, MIN, HR, DAY, YR
from .Electric_Current import Electric_Current, A, MA
from .Temperature import Temperature, DEG_C, DEG_F, DEG_K
from .Amount_of_substance import Amount_of_substance, MOL
from .Luminous_Intensity import Luminous_Intensity, MOL
from .Speed import Speed, KMH, MS, MPH, FPS, LightSpeed
from .Angle import Angle, DEG, MOA, SOA, RAD
from .Resistance import Resistance, OHM, KOHM, MOHM
from .Voltage import Voltage, Volt
from .Volume import Volume, M_3, HL, L, DM_3, DL, CL, ML, CM_3
from .Area import Area, KM_2, HA, ARE, M_2, DM_2, CM_2, MM_2
from .Density import Density, KGM_3, GCM_3
from .Energy import Energy, J, KJ, MJ, GJ
from .Pressure import Pressure, PA
from .Power import Power, W
from .Computer_Memory import Computer_Memory, Bit, Byte, Kb, KB, Mb, MB, Gb, GB, Tb, TB, Pb, PB, Eb, EB, Zb, ZB, Yb, YB
